
from zeep import Client


class CardService:
    def __init__(self, server, user, password, company_id, user_id, process_id=None):
        self.url = 'http://%s/webdesk/CardService?wsdl' % server
        self.client = Client(self.url)
        self.user = user
        self.password = password
        self.company_id = company_id
        self.user_id = user_id
        self.process_id = process_id

    def __get_card_data(self, data):
        """Transforma o dicionário em parâmetros."""

        if bool(data):
            # Obtém classes do webservice.
            try:
                cf_dto_array = self.client.get_type('ns0:cardFieldDtoArray')
                cf_dto = self.client.get_type('ns0:cardFieldDto')
            except:
                cf_dto_array = self.client.get_type('ns1:cardFieldDtoArray')
                cf_dto = self.client.get_type('ns1:cardFieldDto')

            fields = []
            for k, v in data.items():
                field = cf_dto(field=k, value=v)
                fields.append(field)

            # Transforma o dicionário recebido nos objetos do webservice.
            parametros = cf_dto_array(item=fields)
            return parametros
        else:
            return {}

    def update_card_data(self, card_id, card_data):
        """Retorna o byte do arquivo físico de um documento, caso o usuário tenha permissão para acessá-lo."""
        result = self.client.service.updateCardData(self.company_id, self.user, self.password,  card_id,
                                                    cardData=self.__get_card_data(card_data))
        return result




import json
from datetime import timedelta

from WorkflowEngineService import WorkflowEngineService
from DocumentService import DocumentService
from CardService import CardService


class BaseService:
    def __init__(self, url_servidor, username_tbc, senha_tbc, username_ecm, id_processo=None, numero_solicitacao=None,
                 numero_ficha=None, id_empresa=1):
        """Inicia uma instância da classe básica de conexão com o webservice do TOTVS ECM.

        Args:
            url_servidor (int): URL do servidor TBC, incluindo o protocolo, o domínio e a porta, caso necessário.
            username_tbc(str): Username do usuário do TBC.
            senha_tbc(str): Senha do usuário do TBC.
            username_ecm(str): Username do usuário responsável no ECM.
            id_processo(str): Identificador do processo no ECM.
            numero_solicitacao(int): Número da solicitação no ECM.
            numero_ficha(str): Número da ficha relacionada à solicitação no ECM.
            id_empresa(str): Identificador da empresa no ECM.
        """
        self.id_processo = id_processo
        self.numero_solicitacao = numero_solicitacao
        self.numero_ficha = numero_ficha
        self.usuario = username_tbc

        # instâncias dos serviços
        self.__workflowservice = WorkflowEngineService(url_servidor, user=username_tbc, password=senha_tbc,
                                                       company_id=id_empresa, user_id=username_ecm)
        self.__documentservice = DocumentService(url_servidor, user=username_tbc, password=senha_tbc,
                                                 company_id=id_empresa, user_id=username_ecm)
        self.__cardservice = CardService(url_servidor, user=username_tbc, password=senha_tbc, company_id=id_empresa,
                                         user_id=username_ecm)

    @staticmethod
    def __analisar_retorno(data, pkey):
        """Analisa o retorno do ECM em busca de informações."""
        for d in data:
            if d['key'] == pkey:
                return d['value']
        return None

    @property
    def anexos(self):
        """Retorna os anexos do processo. Não inclui os conteúdos por uma questão de otimização."""
        assert self.numero_solicitacao is not None, 'Informe o número da solicitação cujos anexos deseja.'

        attachments = {}

        # Consulta a lista de documentos relacionados ao processo.
        attachments_info = self.__workflowservice.get_attachments(self.numero_solicitacao)
        for attachment in attachments_info:
            attachment_sequence = int(attachment['attachmentSequence'])

            # Exclui o primeiro anexo da lista, que é o formulário do processo.
            if attachment_sequence > 1:
                # Armazena as informações do documento.
                document_id = int(attachment['documentId'])
                colleague_id = attachment['colleagueId']

                # Consulta as informações do documento.
                document_info = self.__documentservice.get_active_document(document_id, colleague_id)[0]
                phisical_file = document_info['phisicalFile']
                version = int(document_info['version'])

                # Consulta o conteúdo do documento.
                # document_content = self.__documentservice.get_document_content(document_id, colleague_id, version,
                #                                                                phisical_file)

                # Adiciona o anexo ao dicionário de anexos.
                attachments[phisical_file] = {
                    'colleague_id': colleague_id,
                    'document_id': document_id,
                    'version': version,
                    # 'document_content': document_content,
                }
        return attachments

    @property
    def atividade_atual(self):
        """Retorna a atividade no qual o processo se encontra atualmente."""
        assert self.numero_solicitacao is not None, 'Informe o número do processo cuja atividade atual deseja.'
        rs = self.__workflowservice.get_all_active_states(self.numero_solicitacao)
        if rs:
            return [int(n) for n in rs]
        else:
            return [-1]

    @property
    def finalizado(self):
        """Indica se o processo foi finalizado."""
        return self.atividade_atual == [-1]

    @property
    def historico(self):
        """Histórico da solicitação."""
        assert self.numero_solicitacao is not None, 'Informe o número do processo cujo histórico deseja.'
        rs = self.__workflowservice.get_histories(self.numero_solicitacao)
        return rs

    @property
    def historico_tratado(self):
        """Retorna o histórico da solicitação de forma tratada."""
        historico = self.historico
        rs = list()
        for etapa in historico:
            for task in etapa.tasks:
                data_hora = task.taskCompletionDate + timedelta(seconds=task.taskCompletionHour) \
                    if task.taskCompletionDate else None
                data_hora_formatada = data_hora.strftime('%d/%m/%Y %H:%M') if task.taskCompletionDate \
                    else u'Tarefa não finalizada'
                rs.append({
                    'data_hora': data_hora,
                    'data_hora_formatada': data_hora_formatada,
                    'proxima_atividade': int(task.choosedSequence),
                    'observacao': task.taskObservation,
                    'colleague_id': task.colleagueId,
                    'texto': task.historCompleteColleague,
                    'nome_responsavel': task.historCompleteColleague.split('\n')[0]
                })
        return rs

    @property
    def responsavel_atual(self):
        """Id do responsável atual da solicitação."""
        return self.historico_tratado[0]['colleague_id']

    def atualizar_formulario(self, dados_formulario):
        """Atualiza o formulário da solicitação.

        Args:
            dados_formulario(dict): Dicionário com os dados a serem atualizados no formulário.

        Returns:
            dict: Resultado da atualização da solicitação.
        """
        assert self.numero_ficha is not None, 'Informe o número da ficha que deseja atualizar.'
        result = self.__cardservice.update_card_data(self.numero_ficha, dados_formulario)
        return result

    def avancar(self, n_atividade, colleague_ids=None, manager_mode=False, observacao=u'Avançado automaticamente'):
        """Avança o processo para uma determinada atividade.

        Args:
            n_atividade(int): Número da atividade para qual a solicitação deve ser avançada.
            colleague_ids(list): Lista de usuários que receberão a solicitação.
            manager_mode(bool): Indica se colaborador esta executando a tarefa como gestor do processo.
            observacao(unicode): Comentários da movimentação.

        Returns:
            dict: Resultado do avanço da solicitação.
        """
        user = colleague_ids if colleague_ids else self.usuario
        assert self.numero_solicitacao is not None, 'Informe o número da solicitação que deseja avançar.'
        result = self.__workflowservice.save_and_send_task_classic(self.numero_solicitacao, n_atividade, [user],
                                                                   observacao, {}, manager_mode=manager_mode)
        return result

    def calcular_prazo(self, data, segundos, prazo, period_id):
        """Calcula o prazo de uma atividade considerando um expediente.

        Args:
            data(str): Data no formato yyy-MM-dd.
            segundos(int): Quantidade de segundos após a meia noite.
            prazo(int): Prazo que será aplicado em horas.
            period_id(str): Código de expediente.

        Returns:
            dict: Resultado do cálculo de prazo.
        """
        prazo = self.__workflowservice.calculate_deadline_hours(data=data, segundos=segundos, prazo=prazo,
                                                                period_id=period_id)
        prazo_str = prazo.__str__().replace('\'', '"')
        prazo_dict = json.loads(prazo_str)
        return prazo_dict

    def cancelar_solicitacao(self, mensagem):
        """Cancela a solicitação.

        Args:
            mensagem(str): Mensagem de cancelamento para o histórico.

        Returns:
            dict: Resultado do cancelamento da solicitação.
        """
        assert self.numero_solicitacao is not None, 'Informe o número do processo que deseja cancelar.'
        result = self.__workflowservice.cancel_instance(self.numero_solicitacao, mensagem)
        return result

    def carregar_solicitacao(self):
        """Carrega as informações da solicitação no objeto do serviço, incluindo número da ficha e ID do processo.

        Returns:
            dict: Dados do formulário da solicitação.
        """
        assert self.numero_solicitacao is not None, 'Informe o número do processo que deseja carregar.'

        # Carrega o número da ficha, que é um dos anexos da solicitação.
        attachments_info = self.__workflowservice.get_attachments(self.numero_solicitacao)
        for attachment in attachments_info:
            attachment_sequence = int(attachment['attachmentSequence'])
            if attachment_sequence == 1:
                self.numero_ficha = int(attachment['documentId'])

        # Carrega as informações do formulário como atributos da solicitação.
        result = self.__workflowservice.get_instance_card_data(self.numero_solicitacao)
        for item in result:
            attribute = item['item'][0]
            value = item['item'][1]
            setattr(self, attribute, value)

            # Carrega o id_processo. Poderia ser feito de outra forma
            # mais elegante, mas assim os editores não apontam erro.
            if attribute == 'WKDef':
                self.id_processo = value

        return result

    def filtrar_historico(self, sequencia, excluir_automaticos=True):
        """Filtra o histórico da solicitação a partir dos identificadores das sequências.

        sequencia(list): Lista de sequências a serem filtradas do histórico.
        excluir_automaticos(bool): Indica se as movimentações automáticas devem ser excluídas do histórico.

        Returns:
            list: Resultado do filtro do histórico.
        """
        rs = [h for h in self.historico_tratado if h['proxima_atividade'] in sequencia]

        # Aplica o filtro caso seja necessário excluir os históricos automáticos.
        if excluir_automaticos:
            rs = [h for h in rs if not h['colleague_id'].startswith('Pool')]

        return rs

    def iniciar_solicitacao(self, dados_formulario, ids_destinatarios, comentarios, anexos=None, completar=True,
                            numero_atividade=0, gestor_processo=False):
        """Inicia uma solicitação no ECM.

        Args:
            dados_formulario(dict): Número da solicitação.
            ids_destinatarios(list):
            comentarios(str): Comentário que fica registrado como
            anexos(dict): Dicionário com nome e conteúdo dos anexos.
            completar(bool): Indica se a tarefa deve ser completada ou apenas salva.
            numero_atividade(int): Número da atividade.
            gestor_processo(bool): Indica se a solicitação está sendo iniciada por um gestor do processo.

        Returns:
            dict: Resultado da inicialização da solicitação.
        """
        assert self.id_processo is not None, 'Informe o ID do processo que deseja iniciar.'
        result = self.__workflowservice.start_process_classic(process_id=self.id_processo,
                                                              colleague_ids=ids_destinatarios,
                                                              card_data=dados_formulario, comments=comentarios,
                                                              attachments=anexos, complete_task=completar,
                                                              choosed_state=numero_atividade,
                                                              manager_mode=gestor_processo)
        iprocess = self.__analisar_retorno(result, 'iProcess')
        if iprocess:
            self.numero_solicitacao = iprocess
            self.numero_ficha = self.__analisar_retorno(result, 'WDNrDocto')
            return result
        else:
            erro = self.__analisar_retorno(result, 'ERROR')
            raise Exception(erro)

    def movimentar(self, origem, destino, observacao='Movimentado automaticamente', **campos_atualizar):
        """Movimenta um processo de uma determinada atividade para outra, verificando antes a atividade de origem e
           possibilitando atualizar campos do formulário, caso necessário para o avanço.

        Args:
            origem(int): Número da atividade de origem, de onde a solicitação vai sair.
            destino(int): Número da atividade de destino, para onde a solicitação vai.
            observacao(str): Comentários da movimentação.
            campos_atualizar(dict): Dados a serem atualizados no formulário no momento da movimentação.

        Returns:
            dict: Resultado da movimentação da solicitação.
        """
        carddata = self.carregar_solicitacao()
        if not self.finalizado:
            if self.atividade_atual[0] == origem:
                # Caso haja campos para atualizar antes de avançar o processo.
                if campos_atualizar:
                    for campo, valor in campos_atualizar.items():
                        if hasattr(self, campo):
                            for obj in carddata:
                                item = obj['item']
                                nome = item[0]
                                if nome == campo:
                                    item[1] = valor
                        else:
                            raise Exception(f'O processo não possui o campo "{campo}".')
                    self.atualizar_formulario(carddata)
                return self.avancar(destino, observacao=observacao, manager_mode=True)
            else:
                raise Exception(f'O processo não está na atividade de origem esperada.')
        else:
            raise Exception('O processo já está finalizado.')




from sys import getsizeof
from zeep import Client


class WorkflowEngineService:
    def __init__(self, server, user, password, company_id, user_id, process_id=None):
        self.url = 'http://%s/webdesk/WorkflowEngineService?wsdl' % server
        self.client = Client(self.url)
        self.user = user
        self.password = password
        self.company_id = company_id
        self.user_id = user_id
        self.process_id = process_id

    def __get_card_data(self, data):
        """Transforma o dicionário em parâmetros."""
        if bool(data):
            # Obtém classes do webservice.
            try:
                kv_dto_array = self.client.get_type('ns0:keyValueDtoArray')
                kv_dto = self.client.get_type('ns0:keyValueDto')
            except:
                kv_dto_array = self.client.get_type('ns1:keyValueDtoArray')
                kv_dto = self.client.get_type('ns1:keyValueDto')

            # Transforma o dicionário recebido nos objetos do webservice.
            fields = []
            for key, value in data.items():
                field = kv_dto(key=key, value=value)
                fields.append(field)
            parametros = kv_dto_array(item=fields)
            return parametros
        else:
            return {}

    def __get_document_array(self, data):
        """Transforma o dicionário com informações dos anexos no tipo Document Array nativo do webservice ."""
        # Instanciamento dos tipos de dados.
        try:
            attachment_type = self.client.get_type('ns0:attachment')
            document_type = self.client.get_type('ns0:processAttachmentDto')
            documents_type = self.client.get_type('ns0:processAttachmentDtoArray')
        except:
            attachment_type = self.client.get_type('ns1:attachment')
            document_type = self.client.get_type('ns1:processAttachmentDto')
            documents_type = self.client.get_type('ns1:processAttachmentDtoArray')

        documents = []

        # Para todos os arquivos presentes na lista de arquivos.
        for file_name, file_info in data.items():
            # Informações do arquivo.
            file_content = file_info['content']
            file_description = file_info['description']

            # Criação do anexo.
            attachment = attachment_type(
                fileName=file_name,
                fileSize=getsizeof(file_content),
                filecontent=file_content,
            )

            # Criação do documento com o anexo.
            document = document_type(
                attachmentSequence=1,
                attachments=[attachment],
                companyId=1,
                deleted=False,
                description=file_description,
                isEdited=False,
                newAttach=True,
                processInstanceId=1,
                fileName=file_name
            )

            # Adiciona o documento à lista de documentos.
            documents.append(document)

        # Cria e retorna o array de documentos.
        document_array = documents_type(item=documents)
        return document_array

    def calculate_deadline_hours(self, data, segundos, prazo, period_id):
        """Calcula um prazo a partir de uma data com base no expediente 
        e feriados cadastrados no produto passando o prazo em horas.

         Args:
            data(str): Data no formato yyy-MM-dd.
            segundos(int): Quantidade de segundos após a meia noite.
            prazo(int): Prazo que será aplicado em horas.
            period_id(str): Código de Expediente.

        Returns:
            str: Objeto DeadLineDto que contem variáveis com a data e hora.
        """
        result = self.client.service.calculateDeadLineHours(self.user, self.password, self.company_id, self.user_id,
                                                            data, segundos, prazo, period_id)
        return result

    def cancel_instance(self, process_instance_id, cancel_text):
        """Cancela uma solicitação.

         Args:
            process_instance_id(int): Número da solicitação.
            cancel_text(str): Comentários do cancelamento.

        Returns:
            str: Mensagem de retorno do cancelamento.
        """
        result = self.client.service.cancelInstance(self.user, self.password, self.company_id, process_instance_id,
                                                    self.user_id, cancel_text)
        return result

    def get_all_active_states(self, process_instance_id):
        """Retorna o número da atividade em que uma solicitação esta.

        Args:
            process_instance_id(int): Número da solicitação.

        Returns:
            list: Número da atividade.
        """
        result = self.client.service.getAllActiveStates(self.user, self.password, self.company_id, self.user_id,
                                                        process_instance_id)
        return result

    def get_attachments(self, process_instance_id):
        """Retorna os anexos de uma solicitação.
        
        Args:
            process_instance_id(int): Número da solicitação.

        Returns:
            list: Lista de anexos.
        """
        result = self.client.service.getAttachments(self.user, self.password, self.company_id, self.user_id,
                                                    process_instance_id)
        return result

    def get_card_value(self, process_instance_id, card_field_name):
        """Retorna o valor de um campo da ficha.

        Args:
            process_instance_id (int): Número da solicitação.
            card_field_name(str): Nome do campo da ficha.

        Returns:
            str: Valor do campo.
        """
        result = self.client.service.getCardValue(self.user, self.password, self.company_id, process_instance_id,
                                                  self.user_id, card_field_name)
        return result

    def get_histories(self, process_instance_id):
        """Retorna lista de históricos de um processo.

        Args:
            process_instance_id(int): Número da solicitação.

        Returns:
            list: Lista com as informações do formulário.
        """
        result = self.client.service.getHistories(self.user, self.password, self.company_id, self.user_id,
                                                  process_instance_id)
        return result

    def get_instance_card_data(self, process_instance_id):
        """Retorna o valor dos campos da ficha de uma solicitação.

        Args:
            process_instance_id(int): Número da solicitação.

        Returns:
            list: Lista com as informações do formulário.
        """
        result = self.client.service.getInstanceCardData(self.user, self.password, self.company_id, self.user_id,
                                                         process_instance_id)
        return result

    def save_and_send_task_classic(self, process_instance_id, choosed_state, colleague_ids, comments, card_data,
                                   manager_mode=False, thread_sequence=0, complete_task=True):
        """Movimenta solicitação para próxima atividade e retorna um array de objeto com chave e valor.

        Args:
            process_instance_id(int): Número da solicitação.
            choosed_state(int): Número da atividade.
            colleague_ids(list): Colaborador que receberá a tarefa.
            comments(unicode): Comentários.
            card_data(dict): Dados da ficha.
            manager_mode(bool): Indica se colaborador esta executando a tarefa como gestor do processo.
            thread_sequence(int): Indica se existe atividade paralela no processo. Se não existir o valor é 0 (zero),
                caso exista, este valor pode ser de 1 a infinito dependendo da quantidade de atividade paralelas
                existentes no processo.
            complete_task(bool): Indica se deve completar a tarefa (true) ou somente salvar (false).

        Returns:
            list: Lista com informações do objeto movimentado.
        """
        result = self.client.service.saveAndSendTaskClassic(self.user, self.password, self.company_id,
                                                            process_instance_id, choosed_state, colleague_ids, comments,
                                                            self.user_id, complete_task, attachments={},
                                                            cardData=self.__get_card_data(card_data), appointment={},
                                                            managerMode=manager_mode, threadSequence=thread_sequence)
        return result

    def start_process_classic(self, process_id, colleague_ids, card_data, comments, attachments=None,
                              complete_task=True, choosed_state=0, manager_mode=False):
        """Inicia uma solicitação e retorna um array de objeto com chave e valor.

        Args:
            process_id(str): Código do processo.
            colleague_ids(list): Colaborador que receberá a tarefa.
            card_data(dict): Dados da ficha dicionarizados.
            comments(str): Comentários.
            attachments(dict): Dicionário com nome e conteúdo lido dos arquivos anexos.
            complete_task(bool): Indica se deve completar a tarefa (True) ou somente salvar (False).
            choosed_state(int): Número da atividade.
            manager_mode(bool): Indica se colaborador esta iniciando a solicitação como gestor do processo.

        Returns:
            list: Lista com informações do objeto criado.
        """
        result = self.client.service.startProcessClassic(self.user, self.password, self.company_id, process_id,
                                                         choosed_state, colleague_ids, comments, self.user_id,
                                                         complete_task, appointment={},
                                                         attachments=self.__get_document_array(attachments),
                                                         cardData=self.__get_card_data(card_data),
                                                         managerMode=manager_mode)
        return result




from setuptools import setup

with open('README.rst', 'r') as fh:
    long_description = fh.read()

setup(name='totvsecm',
      version='1.5.0',
      description='API para webservices do TOTVS ECM.',
      long_description=long_description,
      url='http://github.com/soslaio/totvsecm',
      author='soslaio',
      author_email='rogeriorp@gmail.com',
      license='MIT',
      classifiers=[
            'Development Status :: 4 - Beta',
            'License :: OSI Approved :: MIT License',
            'Programming Language :: Python :: 3.7',
            'Topic :: Software Development :: Libraries :: Python Modules'
      ],
      packages=['totvsecm'],
      install_requires=[
          'zeep',
      ],
      zip_safe=False)



# -*- coding: utf-8 -*-

from zeep import Client


class DocumentService:
    def __init__(self, server, user, password, company_id, user_id, process_id=None):
        self.url = 'http://%s/webdesk/DocumentService?wsdl' % server
        self.client = Client(self.url)
        self.user = user
        self.password = password
        self.company_id = company_id
        self.user_id = user_id
        self.process_id = process_id

    def get_active_document(self, nr_document_id, colleague_id):
        """Retorna um documento ativo.

        Args:
            nr_document_id(int): Número do documento.
            colleague_id(str): Matrícula do colaborador.

        Returns:
            list: Informações do documento.
        """
        result = self.client.service.getActiveDocument(self.user, self.password, self.company_id, nr_document_id,
                                                       colleague_id)
        return result

    def get_document_content(self, nr_document_id, colleague_id, documento_versao, nome_arquivo):
        """Retorna o byte do arquivo físico de um documento, caso o usuário tenha permissão para acessá-lo.

        Args:
            nr_document_id(int): Número do documento.
            colleague_id(str): Matrícula do colaborador.
            documento_versao(int): Número da versão do documento.
            nome_arquivo(str): Nome do arquivo.

        Returns:
            str: Conteúdo do documento.
        """
        result = self.client.service.getDocumentContent(self.user, self.password, self.company_id, nr_document_id,
                                                        colleague_id,
                                                        documento_versao,
                                                        nome_arquivo)
        return result
